/*{{{  File banner*/
/*@(#)=====================================================*\
||@(#)  Project : GPMIMD ESPRIT P5404
||@(#)  Authors : Mark Debbage, Mark Hill, Denis Nicole and Sean Wykes
||@(#)            University of Southampton
||  
||@(#)    Title : Prototype Source File for generated C
||@(#)   System : Occam 2 C
||@(#)  Version : 1.3
||@(#) Filename : Occam2C.c
||@(#)     Date : 3/4/97
\*@(#)====================================================*/
/*}}}*/

/*{{{  Insert SWITCHES*/
#define LIBRARY_UNIT
#define BYTES_PER_WORD 4
#define TERMINATE_ON_ERROR
#define CHECK_CONVERSIONS
#define CHECK_ARRAYS
#define CHECK_RANGES
#define NO_DEBUG
#define FASTGOTOS
#define SOURCE LITTLE

/*}}}*/
/*{{{  defines*/
#define DEFAULT_MEMSIZE_IN_INTS 1000
#define OCCAM_MAIN

/*}}}*/
/*{{{  includes*/
#include <stdio.h>
#include <errno.h>

#if  (defined USESIGTIMER) || (defined USE_NONBLOCKING_SOCKETS)
#include <signal.h>
#include <sys/time.h>
#endif

#ifdef USE_NONBLOCKING_SOCKETS
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef solaris
#include <stropts.h>
#include <sys/file.h>
#include <sys/conf.h>
#endif
#endif

#ifdef CALL_NAG_F90
#include <f90.h>
#endif

#include "Occam2C.h"

#ifdef USE_NONBLOCKING_SOCKETS
  #ifdef solaris
  #define SOCKET_SIGNAL SIGPOLL
  #else
  #define SOCKET_SIGNAL SIGIO
  #endif
#endif
/*}}}*/
/*{{{  Insert HEADER*/
typedef struct SF_P_ss_scrstream_copy_scrcopy tSF_P_ss_scrstream_copy_scrcopy;
void P_ss_scrstream_copy_scrcopy (tSF_P_ss_scrstream_copy_scrcopy *FP);
struct SF_P_ss_scrstream_copy_scrcopy
{
  tHeader _Header;
  void *Chain;
  CHAN *scrn_in_82;
  CHAN *scrn_out_83;
  BOOL going_84;
  union
  {
    struct
    {
      BYTE buffer_85[256];
      INT32 len_86;
    } _S0;
    struct
    {
      BYTE st_char_87;
    } _S1;
    struct
    {
      BYTE st_char_88;
    } _S2;
    struct
    {
      INT32 x_89;
    } _S3;
    struct
    {
      INT32 x_90;
      INT32 y_91;
    } _S4;
  } _U0;
};

/*}}}*/

#if (defined GENPROTO) || !(defined LIBRARY_UNIT)
/*{{{  statics*/
jmp_buf STOPP_env;
int MAIN_ARGC ;
char **MAIN_ARGV;
tTask *QFP[NUM_PRIORITY_LEVELS],*QBP[NUM_PRIORITY_LEVELS],*CURTASK,*QTM; 
static volatile int TimerEvent = false;
static volatile int SocketEvent = false;
tTimer TIMER;
static int process_count;


/*}}}*/
#endif

#ifdef GENPROTO
/*{{{  seterr*/
void seterr(char *M,int L,char *F) {}
/*}}}*/
/*{{{  struct Prototypes*/
struct Prototypes
{
  char *Proto;
  char *Define;
  int Space;
};
/*}}}*/
/*{{{  static struct Prototypes Table[] =*/
static struct Prototypes Table[] =
{
  {"void P_ss_scrstream_copy_scrcopy (tSF_P_ss_scrstream_copy_scrcopy *FP);","P_ss_scrstream_copy_scrcopy",sizeof(tSF_P_ss_scrstream_copy_scrcopy)},

 {NULL,NULL,0}
} ;
/*}}}*/
/*{{{  int main(int argc, char**argv)*/
int main(int argc, char**argv)
{
  char *name;
  FILE *f;
  int fcnt;
  
  CheckPrimitiveSizes();

  if (argc<2)
  {
    name="scrcopy.hdr";
    f=fopen(name,"w");
  }
  else
  {
    name = argv[argc-1];
    if (name[0]=='-')
    {
      name = "stdout";
      f=stdout;
    }
    else
    {
      f=fopen(name,"w");
    }
  }
  fcnt=argc-1;
  
  if (f==NULL)
    { fprintf(stderr,"Error - could not open file %s\n",name);exit(-1); }
  else
  {
    int arg;
    int i;
    fprintf(f,"#ifndef __scrcopy_INCLUDES__\n");
    fprintf(f,"#define __scrcopy_INCLUDES__\n");
    
    for(arg = 1; arg < fcnt; ++arg)
    {
      FILE *fh = fopen(argv[arg],"r");
      if (fh == NULL) { fprintf(stderr,"Error - could not read file %s\n",name); exit(-1); }
      while(!feof(fh))
      {
	char line[256];

	if (fgets(line,256,fh) == NULL) break;
	if (line[0] == '#')
	  if (strncmp(line,"#define WS",10)!=0) continue;
	fputs(line,f);
      }
      fclose(fh);
    }      
      
    i = 0;
    while (Table[i].Proto!=NULL)
    {
      if (Table[i].Define != NULL)
      {
	fprintf(f,"typedef struct SF_%s tSF_%s;\n",Table[i].Define,Table[i].Define);
	fprintf(f,"#define WS_%s (%d)\n",Table[i].Define,Table[i].Space);
      }
      fprintf(f,"%s\n",Table[i].Proto);
      i++;
    }
    fprintf(f,"#endif\n");

    if (f!=stdout) fclose(f);
  }
  return(0);
}     
/*}}}*/
#endif

#ifndef LIBRARY_UNIT
/*{{{  seterr*/
static void DONE_SETERR() {}

void seterr(char *M,int L,char *F)
{
  if (M!=NULL) fputs(M,stderr);
  fprintf(stderr,MSG_EFROL,L,F);
  DONE_SETERR();
  SETERR_PROCEED;
}
/*}}}*/
/*{{{  DEBUGDONE,MARK_SCHEDULE*/
#ifndef NO_DEBUG
void DEBUGDONE(void) {}
void MARK_SCHEDULE(void) {}
#endif

/*}}}*/
/*{{{  timer*/
#ifdef USESIGTIMER
static struct timeval tv_start;
/*{{{  static void TimerSignalHandler(int sig)*/
static void TimerSignalHandler(int sig)
{
  switch(sig)
  {
    case SIGALRM:
      {
	struct timeval tv;
	tTask *task;

	gettimeofday(&tv,NULL);
	TIMER = ((tv.tv_usec - tv_start.tv_usec)>>6) + (1000000>>6) * (tv.tv_sec - tv_start.tv_sec);
	if (QTM != NoTask && AFTER(TIMER,QTM->comms.timer.time))
	  TimerEvent = true;
      }
      break;

    default:
      fprintf(stderr,"Error - Caught signal %d - Exiting\n",sig);
      fflush(stderr);
      exit(20);
  }
}
/*}}}*/
/*{{{  static void Init_Timer()*/
static void Init_Timer()
{
  struct itimerval itv;
#ifndef solaris
  struct sigaction sa;
#endif

  TIMER = 0;
  gettimeofday(&tv_start,NULL);
  TimerSignalHandler(SIGALRM);
  
#ifndef solaris
  memset((char*)&sa,0,sizeof(sa));
  sa.sa_handler = TimerSignalHandler;

#ifdef SA_RESTART
  sa.sa_flags = SA_RESTART;
#endif

  sigaction(SIGALRM,&sa,NULL);
#else
  sigset(SIGALRM,TimerSignalHandler);
#endif

  itv.it_interval.tv_usec = 1000;
  itv.it_interval.tv_sec = 0;
     
  itv.it_value.tv_usec = 1;
  itv.it_value.tv_sec = 0;

  if (setitimer(ITIMER_REAL,&itv,NULL)) 
    fprintf(stderr,"Error - setitimer failed\n");

  QTM = NoTask;
  TimerEvent = false;
}

/*}}}*/

#else
static void Init_Timer() { (void)clock(); TimerEvent = false; }
#endif
/*}}}*/
/*{{{  sockets*/
#ifdef USE_NONBLOCKING_SOCKETS
static int io_limit;
static struct timeval io_zerotime;

static int    io_socket;
static tTask *io_task = NoTask;

/*{{{  void SocketHandler (int sig)*/
void SocketHandler (int sig)
{
  switch (sig)
  {
    case SOCKET_SIGNAL:
    {
      #ifdef MD_DEBUG
      fprintf(stderr,"Caught signal %d\n",sig);
      #endif
      
      SocketEvent = true;
      
      #ifdef solaris
      signal (SOCKET_SIGNAL, SocketHandler);
      #endif
      
      break;
    }
    default:
      fprintf(stderr,"Error - Caught unexpected signal %d - exiting\n",sig);
      fflush(stderr);
      exit(20);
  }  
}
/*}}}*/
 
/*{{{  void WaitSocket (int socket)*/
void WaitSocket (int socket)
{
  fd_set io_set;

  io_socket = socket;

  FD_ZERO (&io_set);
  FD_SET  (socket,&io_set);
  
  errno = 0;
  while (select (io_limit, &io_set, NULL, NULL, &io_zerotime) < 0)
    if (errno != EINTR)
    {
      fprintf(stderr, "Error - Socket select failed\n");
      exit(-1);
    }
    else
      errno = 0;
  
  if (!FD_ISSET(socket,&io_set))
  {
    if (io_task == NoTask)
      io_task = CURTASK;
    else
    {
      fprintf(stderr, "Error - Implementation restriction, only one signaling socket\n");
      exit(-1);
    }
    DeSchedule(TSOCK_p);
  }
}
/*}}}*/

/*{{{  void EnableSocket (int socket)*/
void EnableSocket (int socket)
{
  errno = 0;
  
  #ifdef solaris
  if (ioctl(socket,I_SETSIG,S_RDNORM) < 0)
  {
    fprintf(stderr,"Error - could not set signal handler for socket (errno %d)\n",errno);
    exit(-1);
  }
  #else
  if (fcntl(socket, F_SETOWN, getpid()) < 0)
  {
    fprintf(stderr,"Error - could not set owner on socket (errno %d)\n",errno);
    exit(-1);
  }
  if (fcntl(socket, F_SETFL, FASYNC)    < 0)
  {
    fprintf(stderr,"Error - could not set asynchronous socket (errno %d)\n",errno);
    exit(-1);
  }
  #endif
}
/*}}}*/

/*{{{  void DisableSocket (int socket)*/
void DisableSocket (int socket)
{
  #ifdef sun4
  if (fcntl(socket, F_SETFL, FSYNC) < 0)
  {
    fprintf(stderr,"Error - could not set synchronous socket\n");
    exit(-1);
  }
  #endif
  #ifdef linux
  if (fcntl(socket, F_SETFL, FASYNC) < 0)
  {
    fprintf(stderr,"Error - could not set synchronous socket\n");
    exit(-1);
  }
  #endif
}
/*}}}*/

/*{{{  void ServeSocket (void)*/
void ServeSocket (void)
{
  fd_set io_set;
  int res;

  /* In the general case there's a timing hazard here since a new signal
     could arrive and we can lose it. Really need test-and-set or a
     semaphore. For hostio, however, there's only ever one outstanding 
     communication and this thread is serving it. Hence, no problem here. */

  SocketEvent = false; 

  FD_ZERO (&io_set);
  FD_SET  (io_socket,&io_set);
  
  errno = 0;
  while ((res = select (io_limit, &io_set, NULL, NULL, &io_zerotime)) < 0)
    if (errno != EINTR)
    {
      fprintf(stderr, "Error - Socket select failed\n");
      exit(-1);
    }
    else
      errno = 0;
      
  if (res == 0)
  {
    /*fprintf(stderr,"Could not locate socket signal source, ignoring select\n");*/
    return;
  }
  
  if (!FD_ISSET(io_socket,&io_set))
  {
    fprintf(stderr,"Error - socket is not ready\n");
    exit(-1);
  }

  if (io_task != NoTask)
  {
    ReSchedule(io_task);
    io_task = NoTask;
  }
}
/*}}}*/

/*{{{  void Init_Sockets()*/
void Init_Sockets()
{
  io_limit = ulimit (4,0);

  timerclear (&io_zerotime);

  signal (SOCKET_SIGNAL, SocketHandler);
  #ifdef linux
  {
    struct sigaction action;
    sigaction(SIGIO,NULL,&action);
    action.sa_flags = (action.sa_flags & (~SA_ONESHOT));
    sigaction(SIGIO,&action,NULL);
  }
  #endif
}
/*}}}*/
#else
void WaitSocket (int socket)
{ 
  fprintf(stderr, "Error - No SIGSOCKETS\n"); 
}
#endif
/*}}}*/

/*{{{  static void Init_Scheduler()*/
static void Init_Scheduler()
{
  int i;

  for(i = 0; i < NUM_PRIORITY_LEVELS; ++i)
    QFP[i] = QBP[i] = NoTask;

  CURTASK = NoTask;

  Init_Timer();

  #ifdef USE_NONBLOCKING_SOCKETS
  Init_Sockets();  
  #endif
}
/*}}}*/
/*{{{  static void Scheduler()*/
static void Scheduler()
{
  if (setjmp(STOPP_env)) DeSchedule(0);
  
  while(CURTASK != NoTask || TimerEvent || SocketEvent)
  {
    int pri;

#ifdef USESIGTIMER
    /*{{{  if (TimerEvent)*/
    if (TimerEvent)
    {
    #ifdef DEBUG
	fprintf(stderr,"Timer = %d\n",TIMER);
    #endif
      TimerEvent = false;
      while(QTM != NoTask && AFTER(TIMER,QTM->comms.timer.time))
      {
    #ifdef DEBUG
	{  
	  tTask *task = QTM;
    
	  fprintf(stderr,"Timer Queue contains:");
	  if (task == NoTask) fprintf(stderr,"No Tasks");
	  else while(task != NoTask) { fprintf(stderr,"%d(%d) ",task->TaskId,task->comms.timer.time); task = task->Next; }
	  fprintf(stderr,"\n");
	}  
    #endif
	if (QTM->AltMode == Waiting_p) QTM->AltMode = TReady_p;
	ReSchedule(QTM);
      }
    }
    /*}}}*/
#endif

#ifdef USE_NONBLOCKING_SOCKETS
    if (SocketEvent)
      ServeSocket();
#endif

    for(pri = 0; pri < NUM_PRIORITY_LEVELS; ++pri)
    {
#ifdef DEBUG
  {
    tTask *task = QFP[pri];
    fprintf(stderr,"Task Queue %d contains:",pri);
    if (task == NoTask) fprintf(stderr,"No Tasks");
    else while(task != NoTask) { fprintf(stderr,"%d ",task->TaskId); task = task->Next; }
    fprintf(stderr,"\n");
  }
#endif

      if ((CURTASK=QFP[pri]) != NoTask)
      {
	/*{{{  #ifdef SHOWTASKING*/
	#ifdef SHOWTASKING
		fprintf(stderr,"Scheduler - Current task is %c%d (%s) IP=%d\n",(CURTASK->TaskPriority)?'L':'H',CURTASK->TaskId,((tHeader *)CURTASK->FP)->ProcName,((tHeader *)CURTASK->FP)->IP); fflush(stderr); 
	#endif
	/*}}}*/
	/*{{{  #ifndef NO_DEBUG*/
	#ifndef NO_DEBUG
	MARK_SCHEDULE() ;
	#endif
	/*}}}*/
	CURTASK->Func (CURTASK->FP);
	break;
      }
    }

    /*{{{  catch any signals*/
    #if (defined USESIGTIMER && defined USE_NONBLOCKING_SOCKETS)
	if (CURTASK == NoTask && (QTM != NoTask || io_task != NoTask))
	  while(!TimerEvent && !SocketEvent)
	    pause();
    #else
      #if (defined USESIGTIMER)
	  if (CURTASK == NoTask && QTM != NoTask)
	    while(!TimerEvent)
	      pause();
      #endif
      
      #if (defined USE_NONBLOCKING_SOCKETS)
	  if (CURTASK == NoTask && io_task != NoTask)
	    while(!SocketEvent)
	      pause();
      #endif
    #endif
    /*}}}*/

  }

#ifdef DEBUG
  fprintf(stderr,"Scheduling Queue empty\n");
#endif
}

/*}}}*/
/*{{{  void Start_Process(register tTask *task, register tFuncPtr Func, register void *FP, register char *name,int Priority)*/
void Start_Process(register tTask *task, register tFuncPtr Func, register void *FP, register char *name,int Priority)
{
  static int NextTaskId = 0;
  process_count++;
  
  task->FP = FP;
  task->Func = Func;
  task->state = EXEC_p;
  task->name = name;
  task->Parent = CURTASK;
  task->Children = 0;
  task->Next = NoTask;
  task->AltMode = NotAlting_p;
  task->TaskId = NextTaskId++;
  task->TaskPriority = (CURTASK != NoTask && CURTASK->TaskPriority == HighPriority) ? HighPriority : Priority;
  ((tHeader *)FP)->ProcName = name;
  
  if (Priority < 0 || Priority >= NUM_PRIORITY_LEVELS) SETERR(MSG_PRI);
  
  if (QFP[task->TaskPriority])
    QBP[task->TaskPriority]->Next = task;
  else
    QFP[task->TaskPriority] = task;

  QBP[task->TaskPriority] = task;

  if (CURTASK != NoTask) CURTASK->Children++;
  else CURTASK = task;
  
#ifdef SHOWTASKING
  fprintf(stderr,"  New Task %c%d (%s) FP=%x\n",(task->TaskPriority)?'L':'H',task->TaskId,name,FP);
#endif
}
/*}}}*/
/*{{{  void Start_Placed(register tTask *task, register tFuncPtr Func, register void *FP, register char *name,int Priority)*/
void Start_Placed(register tTask *task, register tFuncPtr Func, register void *FP, register char *name,int Processor)
{
  static int NextTaskId = 0;
  process_count++;

  task->FP = FP;
  task->Func = Func;
  task->state = EXEC_p;
  task->name = name;
  task->Parent = CURTASK;
  task->Children = 0;
  task->Next = NoTask;
  task->AltMode = NotAlting_p;
  task->TaskId = NextTaskId++;
  task->TaskPriority = LowPriority;

  ((tHeader *)FP)->ProcName = name;

  SETERR("PLACED PAR is not supported in this implementation");

  if (QFP[task->TaskPriority])
    QBP[task->TaskPriority]->Next = task;
  else
    QFP[task->TaskPriority] = task;

  QBP[task->TaskPriority] = task;

  if (CURTASK != NoTask) CURTASK->Children++;
  else CURTASK = task;
  
#ifdef SHOWTASKING
  fprintf(stderr,"  New Task %c%d (%s) FP=%x\n",(task->TaskPriority)?'L':'H',task->TaskId,name,FP);
#endif
}
/*}}}*/
/*{{{  void End_Process()*/
void End_Process()
{
  register tTask *task = CURTASK;
  process_count--;
    
  QFP[task->TaskPriority] = task->Next;
  if (QFP[task->TaskPriority] == NoTask) QBP[task->TaskPriority] = NoTask;
  task->Next = NoTask;
  if (task->Parent)
  {
    if (--(task->Parent->Children) == 0) ReSchedule(task->Parent);
  }
}
/*}}}*/

/*{{{  SPOC iserver definitions*/
#ifdef USE_SPOC_ISERVER
extern int CocoPops;
#include "spserver.hdr"

struct SF_P_spoc_iserver_spserver
{
  tHeader _Header;
  void *Chain;
  CHAN *Param0;
  CHAN *Param1;
  char WS[WS_P_spoc_iserver_spserver];
};

typedef struct
{
  tHeader _Header;
  void *Chain;
  CHAN *fs;
  CHAN *ts;
  tSF_P_spoc_iserver_spserver spserver;
} tSF_P_spoc_iserver_caller;

static tSF_P_spoc_iserver_caller spoc_iserver_data;
static tTask spoc_iserver_task;

static tFuncPtr P_spoc_iserver_caller (tSF_P_spoc_iserver_caller *FP)
{
  while(true)
  {
    switch(FP->_Header.IP)
    {
      case  0:
		FP->spserver.Param0=FP->fs;
		FP->spserver.Param1=FP->ts;
		FP->spserver.Chain = NULL;
		CALL(P_spoc_iserver_spserver,&FP->spserver,1,"P_spoc_iserver_spserver");
      case  1:
		ENDP();
    }
  }
}
#endif
/*}}}*/
/*{{{  SPOC aserver definitions*/
#ifdef USE_SPOC_ASERVER
#include "spserver.hdr"

struct SF_P_spoc_aserver_spserver
{
  tHeader _Header;
  void *Chain;
  CHAN *Param0;
  CHAN *Param1;
  char WS[WS_P_spoc_aserver_spserver];
};

typedef struct
{
  tHeader _Header;
  void *Chain;
  CHAN *fs;
  CHAN *ts;
  tSF_P_spoc_aserver_spserver spserver;
} tSF_P_spoc_aserver_caller;

static tSF_P_spoc_aserver_caller spoc_aserver_data;
static tTask spoc_aserver_task;

static tFuncPtr P_spoc_aserver_caller (tSF_P_spoc_aserver_caller *FP)
{
  while(true)
  {
    switch(FP->_Header.IP)
    {
      case  0:
		FP->spserver.Param0=FP->fs;
		FP->spserver.Param1=FP->ts;
		FP->spserver.Chain = NULL;
		CALL(P_spoc_aserver_spserver,&FP->spserver,1,"P_spoc_aserver_spserver");
      case  1:
		ENDP();
    }
  }
}
#endif
/*}}}*/

/*{{{  int main(int argc, char **argv)*/
int main(int argc, char **argv)
{
  void *FP=NULL;
  tSF_ *MAINFP=(tSF_ *)malloc(sizeof(tSF_));
  tTask main_t;
  INT mem_size = (argc>1 && atoi(argv[1])>0)?atoi(argv[1]):DEFAULT_MEMSIZE_IN_INTS;
  INT *mem=(INT*)malloc(mem_size*sizeof(INT));
  CHAN fs,ts;
  
  if (MAINFP==NULL)
  {
    fprintf(stderr,"Error - Malloc of %s bytes for occam workspace failed\n",sizeof(tSF_));
    exit(-1);
  }
  if (mem_size>0 && mem==NULL)
  {
    fprintf(stderr,"Error - Malloc of %s INTs for occam memory array failed\n",mem_size);
    exit(-1);
  }
  (void) clock();
  MAIN_ARGC = argc ;
  MAIN_ARGV = argv ;
  process_count = 0;
 
  CheckPrimitiveSizes();
  INITCH(&fs);
  INITCH(&ts);
  Init_Scheduler();
  
  STARTP(&main_t,,MAINFP,"",LowPriority);

  /*{{{  #ifdef USE_SPOC_ISERVER*/
  #ifdef USE_SPOC_ISERVER
  #ifdef SHOWTASKING
  CocoPops=true;
  #endif
  spoc_iserver_data.fs = &fs;
  spoc_iserver_data.ts = &ts;
  CURTASK = NoTask;  /* needed because don't want this a child of main_t */
  STARTP(&spoc_iserver_task, P_spoc_iserver_caller, &spoc_iserver_data, "P_spoc_iserver_spserver",LowPriority);
  #endif
  /*}}}*/
  /*{{{  #ifdef USE_SPOC_ASERVER*/
  #ifdef USE_SPOC_ASERVER
  spoc_aserver_data.fs = &fs;
  spoc_aserver_data.ts = &ts;
  CURTASK = NoTask;  /* needed because don't want this a child of main_t */
  STARTP(&spoc_aserver_task, P_spoc_aserver_caller, &spoc_aserver_data, "P_spoc_aserver_spserver",LowPriority);
  #endif
  /*}}}*/
/*{{{  #ifdef CALL_NAG_F90*/
#ifdef CALL_NAG_F90
  f90_init(argc,argv);
#endif
/*}}}*/
  
  Scheduler();

/*{{{  #ifdef CALL_NAG_F90*/
#ifdef CALL_NAG_F90
  f90_finish(0);
#endif
/*}}}*/
  if (process_count == 0)
    return 0;
  else
  {
    fprintf(stderr,"Deadlock: %d stopped process%s.\n", process_count,(process_count>1)?"es":"");
    return 1;
  }
}     
/*}}}*/
#endif

#ifndef GENPROTO
/*{{{  Insert DEBUG*/

/*}}}*/
/*{{{  Insert CODE*/
void P_ss_scrstream_copy_scrcopy (tSF_P_ss_scrstream_copy_scrcopy *FP)
#line    5 "scrcopy.occ"
{
#line    5 "scrcopy.occ"
  while(true)
#line    5 "scrcopy.occ"
  {
#line    5 "scrcopy.occ"
    switch(FP->_Header.IP)
#line    5 "scrcopy.occ"
    {
#line    5 "scrcopy.occ"
      CASE(0):  
#line   24 "scrcopy.occ"
        FP->going_84 = true; 
#line   25 "scrcopy.occ"
        GOTO(1);
#line   25 "scrcopy.occ"
      CASE(2):  
#line   27 "scrcopy.occ"
        INPUT1(FP->scrn_in_82,&FP->_Header.Temp.VBYTE,3);
#line   27 "scrcopy.occ"
      CASE(3):  
#line   27 "scrcopy.occ"
        switch(FP->_Header.Temp.VBYTE)
#line   27 "scrcopy.occ"
        {
#line   27 "scrcopy.occ"
          case 8: GOTO (5);
#line   27 "scrcopy.occ"
          case 18: GOTO (6);
#line   27 "scrcopy.occ"
          case 6: GOTO (7);
#line   27 "scrcopy.occ"
          case 19: GOTO (8);
#line   27 "scrcopy.occ"
          case 5: GOTO (9);
#line   27 "scrcopy.occ"
          case 0: GOTO (10);
#line   27 "scrcopy.occ"
          case 1: GOTO (11);
#line   27 "scrcopy.occ"
          case 2: GOTO (12);
#line   27 "scrcopy.occ"
          case 3: GOTO (13);
#line   27 "scrcopy.occ"
          case 4: GOTO (14);
#line   27 "scrcopy.occ"
          case 7: GOTO (15);
#line   27 "scrcopy.occ"
          case 9: GOTO (16);
#line   27 "scrcopy.occ"
          case 10: GOTO (17);
#line   27 "scrcopy.occ"
          case 11: GOTO (18);
#line   27 "scrcopy.occ"
          case 12: GOTO (19);
#line   27 "scrcopy.occ"
          case 13: GOTO (20);
#line   27 "scrcopy.occ"
          case 15: GOTO (21);
#line   27 "scrcopy.occ"
          case 16: GOTO (22);
#line   27 "scrcopy.occ"
          case 17: GOTO (23);
#line   27 "scrcopy.occ"
          case 20: GOTO (24);
#line   27 "scrcopy.occ"
          case 21: GOTO (25);
#line   27 "scrcopy.occ"
          case 22: GOTO (26);
#line   27 "scrcopy.occ"
          case 23: GOTO (27);
#line   27 "scrcopy.occ"
          case 24: GOTO (28);
#line   27 "scrcopy.occ"
          default: SETERR(MSG_CASEIN);
#line   27 "scrcopy.occ"
        }
#line   27 "scrcopy.occ"
        break;
#line   27 "scrcopy.occ"
      CASE(5):  
#line   30 "scrcopy.occ"
        INPUT4(FP->scrn_in_82,&FP->_U0._S0.len_86,29);
#line   30 "scrcopy.occ"
      CASE(29):  
#line   30 "scrcopy.occ"
        CHK(FP->_U0._S0.len_86<=256,MSG_CITL);
#line   30 "scrcopy.occ"
        INPUT(FP->scrn_in_82,FP->_U0._S0.buffer_85,(FP->_U0._S0.len_86)*(sizeof(BYTE)),30);
#line   30 "scrcopy.occ"
      CASE(30):  
#line   31 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 8;
#line   31 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,31);
#line   31 "scrcopy.occ"
      CASE(31):  
#line   31 "scrcopy.occ"
        OUTPUT4(FP->scrn_out_83,&FP->_U0._S0.len_86,32);
#line   31 "scrcopy.occ"
      CASE(32):  
#line   31 "scrcopy.occ"
        OUTPUT(FP->scrn_out_83,FP->_U0._S0.buffer_85,(FP->_U0._S0.len_86)*(sizeof(BYTE)),33);
#line   31 "scrcopy.occ"
      CASE(33):  
#line   31 "scrcopy.occ"
        GOTO(4);
#line   31 "scrcopy.occ"
      CASE(6):  
#line   33 "scrcopy.occ"
        INPUT1(FP->scrn_in_82,&FP->_U0._S1.st_char_87,34);
#line   33 "scrcopy.occ"
      CASE(34):  
#line   34 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 18;
#line   34 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,35);
#line   34 "scrcopy.occ"
      CASE(35):  
#line   34 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_U0._S1.st_char_87,36);
#line   34 "scrcopy.occ"
      CASE(36):  
#line   34 "scrcopy.occ"
        GOTO(4);
#line   34 "scrcopy.occ"
      CASE(7):  
#line   36 "scrcopy.occ"
        INPUT1(FP->scrn_in_82,&FP->_U0._S2.st_char_88,37);
#line   36 "scrcopy.occ"
      CASE(37):  
#line   37 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 6;
#line   37 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,38);
#line   37 "scrcopy.occ"
      CASE(38):  
#line   37 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_U0._S2.st_char_88,39);
#line   37 "scrcopy.occ"
      CASE(39):  
#line   37 "scrcopy.occ"
        GOTO(4);
#line   37 "scrcopy.occ"
      CASE(8):  
#line   39 "scrcopy.occ"
        INPUT4(FP->scrn_in_82,&FP->_U0._S3.x_89,40);
#line   39 "scrcopy.occ"
      CASE(40):  
#line   40 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 19;
#line   40 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,41);
#line   40 "scrcopy.occ"
      CASE(41):  
#line   40 "scrcopy.occ"
        OUTPUT4(FP->scrn_out_83,&FP->_U0._S3.x_89,42);
#line   40 "scrcopy.occ"
      CASE(42):  
#line   40 "scrcopy.occ"
        GOTO(4);
#line   40 "scrcopy.occ"
      CASE(9):  
#line   42 "scrcopy.occ"
        INPUT4(FP->scrn_in_82,&FP->_U0._S4.x_90,43);
#line   42 "scrcopy.occ"
      CASE(43):  
#line   42 "scrcopy.occ"
        INPUT4(FP->scrn_in_82,&FP->_U0._S4.y_91,44);
#line   42 "scrcopy.occ"
      CASE(44):  
#line   43 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 5;
#line   43 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,45);
#line   43 "scrcopy.occ"
      CASE(45):  
#line   43 "scrcopy.occ"
        OUTPUT4(FP->scrn_out_83,&FP->_U0._S4.x_90,46);
#line   43 "scrcopy.occ"
      CASE(46):  
#line   43 "scrcopy.occ"
        OUTPUT4(FP->scrn_out_83,&FP->_U0._S4.y_91,47);
#line   43 "scrcopy.occ"
      CASE(47):  
#line   43 "scrcopy.occ"
        GOTO(4);
#line   43 "scrcopy.occ"
      CASE(10):  
#line   45 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 0;
#line   45 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,48);
#line   45 "scrcopy.occ"
      CASE(48):  
#line   45 "scrcopy.occ"
        GOTO(4);
#line   45 "scrcopy.occ"
      CASE(11):  
#line   47 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 1;
#line   47 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,49);
#line   47 "scrcopy.occ"
      CASE(49):  
#line   47 "scrcopy.occ"
        GOTO(4);
#line   47 "scrcopy.occ"
      CASE(12):  
#line   49 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 2;
#line   49 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,50);
#line   49 "scrcopy.occ"
      CASE(50):  
#line   49 "scrcopy.occ"
        GOTO(4);
#line   49 "scrcopy.occ"
      CASE(13):  
#line   51 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 3;
#line   51 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,51);
#line   51 "scrcopy.occ"
      CASE(51):  
#line   51 "scrcopy.occ"
        GOTO(4);
#line   51 "scrcopy.occ"
      CASE(14):  
#line   53 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 4;
#line   53 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,52);
#line   53 "scrcopy.occ"
      CASE(52):  
#line   53 "scrcopy.occ"
        GOTO(4);
#line   53 "scrcopy.occ"
      CASE(15):  
#line   55 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 7;
#line   55 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,53);
#line   55 "scrcopy.occ"
      CASE(53):  
#line   55 "scrcopy.occ"
        GOTO(4);
#line   55 "scrcopy.occ"
      CASE(16):  
#line   57 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 9;
#line   57 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,54);
#line   57 "scrcopy.occ"
      CASE(54):  
#line   57 "scrcopy.occ"
        GOTO(4);
#line   57 "scrcopy.occ"
      CASE(17):  
#line   59 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 10;
#line   59 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,55);
#line   59 "scrcopy.occ"
      CASE(55):  
#line   59 "scrcopy.occ"
        GOTO(4);
#line   59 "scrcopy.occ"
      CASE(18):  
#line   61 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 11;
#line   61 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,56);
#line   61 "scrcopy.occ"
      CASE(56):  
#line   61 "scrcopy.occ"
        GOTO(4);
#line   61 "scrcopy.occ"
      CASE(19):  
#line   63 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 12;
#line   63 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,57);
#line   63 "scrcopy.occ"
      CASE(57):  
#line   63 "scrcopy.occ"
        GOTO(4);
#line   63 "scrcopy.occ"
      CASE(20):  
#line   65 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 13;
#line   65 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,58);
#line   65 "scrcopy.occ"
      CASE(58):  
#line   65 "scrcopy.occ"
        GOTO(4);
#line   65 "scrcopy.occ"
      CASE(21):  
#line   67 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 15;
#line   67 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,59);
#line   67 "scrcopy.occ"
      CASE(59):  
#line   67 "scrcopy.occ"
        GOTO(4);
#line   67 "scrcopy.occ"
      CASE(22):  
#line   69 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 16;
#line   69 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,60);
#line   69 "scrcopy.occ"
      CASE(60):  
#line   69 "scrcopy.occ"
        GOTO(4);
#line   69 "scrcopy.occ"
      CASE(23):  
#line   71 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 17;
#line   71 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,61);
#line   71 "scrcopy.occ"
      CASE(61):  
#line   71 "scrcopy.occ"
        GOTO(4);
#line   71 "scrcopy.occ"
      CASE(24):  
#line   73 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 20;
#line   73 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,62);
#line   73 "scrcopy.occ"
      CASE(62):  
#line   73 "scrcopy.occ"
        GOTO(4);
#line   73 "scrcopy.occ"
      CASE(25):  
#line   75 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 21;
#line   75 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,63);
#line   75 "scrcopy.occ"
      CASE(63):  
#line   75 "scrcopy.occ"
        GOTO(4);
#line   75 "scrcopy.occ"
      CASE(26):  
#line   77 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 22;
#line   77 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,64);
#line   77 "scrcopy.occ"
      CASE(64):  
#line   77 "scrcopy.occ"
        GOTO(4);
#line   77 "scrcopy.occ"
      CASE(27):  
#line   79 "scrcopy.occ"
        FP->_Header.Temp.VBYTE = 23;
#line   79 "scrcopy.occ"
        OUTPUT1(FP->scrn_out_83,&FP->_Header.Temp.VBYTE,65);
#line   79 "scrcopy.occ"
      CASE(65):  
#line   79 "scrcopy.occ"
        GOTO(4);
#line   79 "scrcopy.occ"
      CASE(28):  
#line   81 "scrcopy.occ"
        FP->going_84 = false; 
#line   81 "scrcopy.occ"
        GOTO(4);
#line   81 "scrcopy.occ"
      CASE(4):  
#line   25 "scrcopy.occ"
      CASE(1):  
#line   25 "scrcopy.occ"
        if (FP->going_84)
#line   25 "scrcopy.occ"
        {
#line   25 "scrcopy.occ"
          GOTO(2);
#line   25 "scrcopy.occ"
        }
#line   25 "scrcopy.occ"
        RETURN();
#line   25 "scrcopy.occ"
      default: SETERR(MSG_IP);
#line   25 "scrcopy.occ"
    }
#line   25 "scrcopy.occ"
  }
#line   25 "scrcopy.occ"
}


/*}}}*/
#endif
